/*

File:   c_utils.c
Purpose: utility library

Copyright Fabrice Nicol <fabnicol@users.sourceforge.net>, 2008

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <stdlib.h>
#include <stdio.h>
#ifdef __WIN32__
#include <io.h>
#endif
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#include <stdarg.h>
#include <sys/time.h>

#ifndef __MINGW32__
#include <sys/resource.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <string.h>

#include <time.h>
#include "c_utils.h"


globalData globals;


void pause_dos_type()
{
  do
    {
      char reply;
      fprintf(stderr, "\n%s\n", "Press key to continue...");
      scanf("%c", &reply);
      fflush(stdin);
    }
  while (0);
}


/*********************************************************************************************************
 * function: rm_empty_backup_dir
 *   securely removes empty backup directories, doing security tests.
 **********************************************************************************************************/


int rm_empty_backup_dir(const char * rep)
{
  DIR *dir;
  if (((dir=opendir(rep)) != NULL) &&   (rmdir(rep) == -1))
    return(0);
  else
    return(1);

  closedir(dir) ;
  errno=0 ;
  return(0);
}

/*********************************************************************************************************
 * function: clean_exit
 *   logs time; flushes all streams;  erase empty backup dirs; closes log;
 *   pauses before exiting, with different behaviour on Linux and Windows.
 **********************************************************************************************************/

void clean_exit(int message, const char *default_directory)
{

  //COMPUTE_EXECTIME
  fflush(stdout);

  rm_empty_backup_dir(default_directory);

  if (globals.logfile)
    {
      fflush(globals.journal);
      fclose(globals.journal);
    }
  if (globals.end_pause) pause_dos_type();
  pause_dos_type();
  exit(message);
}

/****************************************************************
* function: mkdir_p
*   Creates directories.
****************************************************************/

int secure_mkdir (const char *path, mode_t mode, const char* default_directory)
{

  int i=0, len;
  len = strlen (path);

  // requires std=c99


  if  ((len<=1) && (globals.debugging))
  {
  foutput("%s\n","[ERR]  Path length could not be allocated by secure_mkdir:\n       Your compiler may not be C99-compliant");
  foutput("       path=%s, length=%d\n",path,len);
  clean_exit(EXIT_FAILURE,default_directory);
  }

  char d[len+1];
  memset(d, 0, len+1);

  memmove(d, path, len+1);

  if (d == NULL)
    {
      perror("[MSG] Error: could not allocate directory path string");
      pause_dos_type();
      clean_exit(EXIT_FAILURE, default_directory);
    }

  errno=0;
  for (i = 1; i < len; i++)
    {
      if (('/' == d[i]) || ('\\' == d[i]))
        {
          d[i] = '\0';

          if ((MKDIR(d, mode) == -1) && (EEXIST != errno))

           {
              foutput( "Impossible to create directory '%s'\n", d);
			  foutput("%s", "Backup directories will be used.\n " );
			  pause_dos_type();
              exit(EXIT_FAILURE);

            }
          d[i] = '/';


        }

    }
  // loop stops before end of string as dirpaths can optionally end in '/' under *nix

  if  (MKDIR(path, mode) == -1)
    {
      if (EEXIST == errno)
        {
          errno=0;
          foutput("[MSG]  Output directory '%s' already created\n", path);
          return(0);
        }
      foutput("[ERR]  Impossible to create directory '%s'\n", path);
      foutput("       permission was: %d\n       %s\n", mode, strerror(errno));
      foutput( "%s", "       Backup directories will be used.\n");
      errno=0;
      pause_dos_type();
      exit(EXIT_FAILURE);
    }

  return(errno);
}


extern struct rusage nothing, start, end;

void starter()
{
#ifndef __MINGW32__
  getrusage(RUSAGE_SELF, &start);
  getrusage(RUSAGE_SELF, &nothing);
  nothing.ru_utime.tv_sec -= start.ru_utime.tv_sec;
  nothing.ru_stime.tv_sec -= start.ru_stime.tv_sec;
#endif
  return;
}

void print_commandline(int argc, char * const argv[])
{
  int i=0;

  foutput("\n%s \n", "[INF]  Running:");

  for (i=0; i < argc ; i++)
    foutput(" %s  ",  argv[i]);
  foutput("%s", "\n\t");

}

void print_time()
{
  char outstr[200];
  time_t t;
  struct tm *tmp;

  t = time(NULL);
  tmp = localtime(&t);
  if (tmp == NULL)
    {
      perror("localtime");
      exit(EXIT_FAILURE);
    }

  if (strftime(outstr, sizeof(outstr), "%c", tmp) == 0)
    {
      foutput("%s\n", "strftime returned 0");
      exit(EXIT_FAILURE);
    }

  foutput("\nCurrent time is: %s", outstr);
}

void change_directory(const char * filename)
{
  if (chdir(filename) == -1)
    {
      if (errno == ENOTDIR)
        foutput("[ERR]  %s is not a directory", filename);
      foutput("[ERR]  input file %s does not comply with chdir  specifications\n.", filename);
      pause_dos_type();
      exit(EXIT_FAILURE);
    }
}

int copy_directory(const char* src, const char* dest, mode_t mode)
{

  DIR *dir_src;

  struct stat buf;
  struct dirent *f;
  char path[BUFSIZ];

  stat(dest, &buf);

  foutput("%c", '\n');

  if (globals.debugging)  foutput("%s%s\n", "[INF]  Creating dir=", dest);
  make_directory(dest, mode, "/usr/local/dvda");

  if (globals.debugging)   foutput("[INF]  Copying in %s ...\n", src);
  change_directory(src);

  dir_src=opendir(".");

  while ((f=readdir(dir_src)) != NULL)
    {
      if (f->d_name[0] == '.') continue;
      STRING_WRITE(path, "%s%c%s", dest, '/', f->d_name)
      if (stat(f->d_name, &buf) == -1)
        {
          perror("[ERR] stat ");
          exit(EXIT_SUCCESS);
        }

      if (S_IFDIR & buf.st_mode)
        {

          if (globals.debugging) foutput("%s %s %s %s\n", "[INF]  Copying dir=", f->d_name, " to=", path);

          errno=copy_directory(f->d_name, path, mode);
          continue;
        }
      if (S_IFREG & buf.st_mode)
        {
          if (globals.debugging) foutput("%s%s to= %s\n", "[INF]  Copying file=", f->d_name, path);
          FILE *infile, *outfile;
          infile=fopen(f->d_name, "rb");
          outfile=fopen(path, "wb+");

          if (infile == NULL)
          {
              foutput("[ERR] %s did not open\n", f->d_name);
              exit(EXIT_FAILURE);
          }
          if (outfile == NULL)
          {
              foutput("[ERR] %s did not open\n", path);
              exit(EXIT_FAILURE);
          }
          errno=copy_file(infile, outfile);
          fclose(infile);
          fclose(outfile);
        }
      /* does not copy other types of files(symlink, sockets etc). */

      else continue;
    }
  change_directory("../");
  if (globals.debugging)   foutput("%s", "[INF]  Done. Backtracking... \n\n");
  closedir(dir_src);
  return(errno);
}

int copy_file(FILE *infile, FILE *outfile)
{

  char buf[BUFSIZ];
  size_t chunk;
  clearerr(infile);
  clearerr(outfile);
  for(;;)
    {
      chunk=fread(buf, sizeof(char), BUFSIZ, infile);
      errno=0;
      if ( errno )
      {
      fprintf(stderr, "[ERR] %s\n", strerror(errno));
      return(errno);
      }
      if (feof(infile)) break;

      fwrite(buf,  sizeof(char), chunk , outfile);
      if (ferror(outfile))
        {
          fprintf(stderr, "Write error\n");
          return(-1);
        }
    }
  return(errno);
}

int get_endianness()
{
  long i=1;
  const char *p=(const char *) &i;
  if (p[0] == 1) return LITTLE_ENDIAN;
  return BIG_ENDIAN;
}





