/*

File:   c_utils.h
Purpose: utility library header

libc_utils.a utility library

Copyright Fabrice Nicol <fabnicol@users.sourceforge.net>, 2008

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/



#ifndef C_UTILS_H_INCLUDED
#define C_UTILS_H_INCLUDED
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

/* Prototypes */

void clean_exit(int message, const char* default_directory);
void help();
void starter();
void print_time();
int rm_empty_backup_dir(const char * rep);
int secure_mkdir ( const char *path, mode_t mode, const char* default_directory);
void print_commandline(int argc, char * const argv[]);
void change_directory(const char * filename);
int copy_file(FILE *infile, FILE* outfile);
int copy_directory(const char* src, const char* dest, mode_t mode);
void pause_dos_type();
int get_endianness();


/* Structures */

typedef struct
  {
    char  *settingsfile;
    char  *logfile;
    char  *indir;
    char  *outdir;
    char  *linkdir;
  } defaults ;


typedef struct
  {
    _Bool silence;
    _Bool logfile;
    _Bool videozone;
    _Bool videolinking;
    _Bool end_pause;
    _Bool debugging;
    FILE *journal;
    defaults settings;

  } globalData ;


/* sets the size of character-type buffers (command-line parsing etc). */


#define Min(X ,Y)    ((X) <= (Y))  ? (X):(Y)
#define MAX(X ,Y)    ((X) <= (Y))  ? (Y):(X)

#ifndef FLOOR_CHAR_BUFSIZ
#define FLOOR_CHAR_BUFSIZ 512
#endif

#ifndef CHAR_BUFSIZ
// multiple of 8 ensures better alignment of structure
#define CHAR_BUFSIZ    (MAX(FLOOR_CHAR_BUFSIZ, BUFSIZ/(8*10)* 8))
#endif


#ifndef LITTLE_ENDIAN
#define LITTLE_ENDIAN  1
#define BIG_ENDIAN  0
#endif

#define ERR_STRING_LENGTH   "ERR: string was truncated, maximum length is %d"

/* MACROS*/

#define foutput(X, ...)   do { if (!globals.silence) printf(X, __VA_ARGS__);\
							   if (!globals.logfile) break;\
							   fprintf(globals.journal, X, __VA_ARGS__);} while(0)

#ifdef __WIN32__
#define	 make_directory(X, Y, Z) secure_mkdir(X, 0, Z)
#define   MKDIR(X, Y) mkdir(X)
#else
#define   MKDIR(X, Y) mkdir(X, Y)
#define make_directory(X, Y, Z)  secure_mkdir(X, Y, Z)
#endif


/* STRING_WRITE is devised for strings allocated by arrays */

#define STRING_WRITE(X,Z,...)	 do { int chres, Y;\
													 Y=sizeof(X); \
													 if    ( ( chres = snprintf(X, Y , Z, __VA_ARGS__) ) >=  Y )  \
														   foutput("\n"ERR_STRING_LENGTH"\n", Y);\
														   else   if (chres < 0 ) foutput( "\n[ERR] Error message:  %s\nCheck source code %s, line %d",  strerror(errno), __FILE__, __LINE__); } while(0);


#define STRING_WRITE_CHAR_BUFSIZ(X,Z,...)	 do { int chres;\
													 if    ( ( chres = snprintf(X, CHAR_BUFSIZ*sizeof(char) , Z, __VA_ARGS__) ) >=  CHAR_BUFSIZ *sizeof(char))  \
														   foutput("\n"ERR_STRING_LENGTH"\n", CHAR_BUFSIZ);\
														   else   if (chres < 0 ) foutput( "\n[ERR] Error message:  %s\nCheck source code %s, line %d",  strerror(errno), __FILE__, __LINE__); } while(0);
/* ERROR MANAGEMENT

    Error management conventions:
       EXIT_SUCCESS is returned on errors caused by ill-formed user input.
       EXIT_FAILURE is reserved for internal errors (segfaults, etc.)
	EXIT_ON_ERROR is a macro for uncommented EXIT_FAILURE
	EXIT_ON_ERROR_VERBOSE  is a macro for commented EXIT_FAILURE (one argument string)
	in other contexts (EXIT_SUCCESS, complex EXIT_FAILURE... )  clean_exit is used, see auxiliaray.c
*/


#define EXIT_ON_ERROR(Y)  if (errno) do {   foutput( "\n%s\n       ::%d, %s\n", strerror(errno), __LINE__, __FILE__);\
															clean_exit(EXIT_FAILURE, Y); } while(0);

#define EXIT_ON_ERROR_VERBOSE(X, Y)  if (errno) do {   foutput( "\n%s\n%s\n       ::%d, %s\n", X, strerror(errno), __LINE__, __FILE__);\
																			  clean_exit(EXIT_FAILURE, Y);  } while(0);


#endif // C_UTILS_H_INCLUDED
