/*

File:    audio.c
Purpose: Deal with reading the input audio files.

dvda-author  - Author a DVD-Audio DVD

(C) Dave Chapman <dave@dchapman.com> 2005

The latest version can be found at http://dvd-audio.sourceforge.net

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "audio2.h"
#include "stream_decoder.h"
#include "dvda-author.h"
#include "c_utils.h"

extern globalData globals;

void flac_metadata_callback(const FLAC__StreamDecoder *dec, const FLAC__StreamMetadata *meta, void *data)
{
  fileinfo_t *info = (fileinfo_t *) data;

  if (meta->type==FLAC__METADATA_TYPE_STREAMINFO)
    {
      info->bitspersample=meta->data.stream_info.bits_per_sample;
      info->samplerate=meta->data.stream_info.sample_rate;
      info->channels=meta->data.stream_info.channels;
      info->numsamples=meta->data.stream_info.total_samples;
    }
}

FLAC__StreamDecoderWriteStatus flac_null_write_callback(const FLAC__StreamDecoder *dec,
    const FLAC__Frame *frame,
    const FLAC__int32 * const buf[],
    void *data)
{
  return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

FLAC__StreamDecoderWriteStatus flac_write_callback(const FLAC__StreamDecoder *dec,
    const FLAC__Frame *frame,
    const FLAC__int32 * const buf[],
    void *data)
{
  fileinfo_t *info = (fileinfo_t*) data;

  int c_samp, c_chan, d_samp;
  uint32_t data_size = frame->header.blocksize * frame->header.channels * (info->bitspersample / 8);
  uint32_t samples = frame->header.blocksize;
  uint32_t i;

  if ((info->audio->n+data_size) > sizeof(info->audio->buf))
    EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Internal error - FLAC buffer overflown")


    // Store data in interim buffer in WAV format - i.e. Little-endian interleaved samples
    i=info->audio->n;
  for (c_samp = d_samp = 0; c_samp < samples; c_samp++)
    {
      for (c_chan = 0; c_chan < frame->header.channels; c_chan++, d_samp++)
        {
          info->audio->buf[i++]=(buf[c_chan][c_samp]&0xff);
          info->audio->buf[i++]=(buf[c_chan][c_samp]&0xff00)>>8;
          if (info->bitspersample==24)
            {
              info->audio->buf[i++]=(buf[c_chan][c_samp]&0xff0000)>>16;
            }
        }
    }
  info->audio->n=i;

  return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}



void flac_error_callback(const FLAC__StreamDecoder *dec,
                         FLAC__StreamDecoderErrorStatus status, void *data)
{
  foutput("%s", "[ERR]  FLAC error callback called.\n");
}



void calc_info(fileinfo_t* info)
{

  info->bytespersecond=(info->samplerate*info->bitspersample*info->channels)/8;
  switch (info->bitspersample)
    {
    case 16:
      info->numsamples=info->numbytes/(info->channels*2);
      info->sampleunitsize=2*info->channels;
      break;
    case 20:
      info->numsamples=(info->numbytes*2)/(info->channels*5);
      info->sampleunitsize=5*info->channels;    // Fix: what is correct sampleunitsize?
      break;
    case 24:
      info->numsamples=info->numbytes/(info->channels*3);
      info->sampleunitsize=6*info->channels;
      break;
    default:
      foutput("Unsupported samplerate: %d\n",info->bitspersample);
      clean_exit(EXIT_SUCCESS, DEFAULT);
    }

  if ((info->samplerate==48000) || (info->samplerate==44100))
    {
      info->bytesperframe=5*info->channels*info->bitspersample;
    }
  else if ((info->samplerate==96000) || (info->samplerate==88200))
    {
      info->bytesperframe=10*info->channels*info->bitspersample;
    }
  else if ((info->samplerate==192000) || (info->samplerate==176400))
    {
      info->bytesperframe=20*info->channels*info->bitspersample;
    }

  info->numsamples=info->numbytes/(info->channels*info->bitspersample/8);

  if ((info->bitspersample==24) && (info->channels==2))
    {
      if ((info->numsamples%2)==1)
        {
          foutput("[WAR]  Padding file %s with 1 sample (6 bytes)\n",info->filename);
          fflush(stdout);
          /* PATCH: numsamples and numbytes should not comprise the padding */
          info->numsamples--;
          info->numbytes-=6;        }
    }

  info->PTS_length=(90000*info->numsamples)/info->samplerate;
}

uint64_t read_file_size(FILE * fp)
{
uint64_t size = 0;

 if (fp)
 {
    /* get size */
    fseek(fp,0L,SEEK_END);
    size = (uint64_t)ftell (fp);
    rewind(fp);
  }
  return size;
}


int flac_getinfo(fileinfo_t* info)
{
  FLAC__StreamDecoder* flac;
  FLAC__StreamDecoderInitStatus  result;

  flac=FLAC__stream_decoder_new();

  if (flac==NULL)
    EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Fatal error - could not create FLAC decoder\n")

    /* Transition from the FLAC 1.1.2 syntax: legacy code recalled below
    *
    *	FLAC__file_decoder_set_filename(flac,info->filename);
    *	FLAC__file_decoder_set_client_data(flac,(void*)info);
    *	FLAC__file_decoder_set_write_callback(flac,flac_null_write_callback);
    *	FLAC__file_decoder_set_error_callback(flac,flac_error_callback);
    *	FLAC__file_decoder_set_metadata_callback(flac,flac_metadata_callback);
    *
    * end of legacy code */

    result=/*FLAC__StreamDecoderInitStatus*/ FLAC__stream_decoder_init_file  	(
          /*FLAC__StreamDecoder *  */ 	 flac,
          /*FILE * */ 	info->filename,
          /*FLAC__StreamDecoderWriteCallback */ 	flac_null_write_callback,
          /*FLAC__StreamDecoderMetadataCallback */ 	flac_metadata_callback,
          /*FLAC__StreamDecoderErrorCallback  */	flac_error_callback,
          (void *) info
        );

  if (result!=FLAC__STREAM_DECODER_INIT_STATUS_OK)
    {

      FLAC__stream_decoder_delete(flac);
      EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Failed to initialise FLAC decoder\n");
    }

  if (!FLAC__stream_decoder_process_until_end_of_metadata(flac))
    {

      FLAC__stream_decoder_delete(flac);
      EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Failed to read metadata from FLAC file\n");
    }
  FLAC__stream_decoder_finish(flac);

  if (((info->bitspersample!=16) && (info->bitspersample!=24)) || (info->channels > 2))
    {
      return(1);
    }

  info->type=AFMT_FLAC;
  info->numbytes=info->numsamples*info->channels*(info->bitspersample/8);
  calc_info(info);
  FLAC__stream_decoder_delete(flac);
  return(0);
}

int wav_getinfo(fileinfo_t* info)
{
  unsigned char header[44]; // Assume 44-byte canonical headers for now...
  FILE* fp;

  fp=fopen(info->filename,"rb");
  if (!fp)
    {
      return(1);
    }
    /* PATCH: real size on disc is needed */
  info->file_size = read_file_size(fp);
  fread(header,44,1,fp);
  fclose(fp);

  if ((memcmp(header,"RIFF",4)!=0) || (memcmp(&header[8],"WAVEfmt",7)!=0))
    {
      return(1);
    }

  info->type=AFMT_WAVE;

  // FIX: Correctly parse WAV header.  For now, assume canonical 44-byte header
  // Using fixwav here

  info->samplerate=header[24]|(header[25]<<8)|(header[26]<<16)|(header[27]<<24);
  info->bitspersample=header[34];
  info->channels=header[22];
  info->numbytes=(header[40]|(header[41]<<8)|(header[42]<<16)|(header[43]<<24));

  if (((info->bitspersample!=16) && (info->bitspersample!=24)) || (info->channels > 2))
    {
      return(2);
    }

  if(info->numbytes > info->file_size - 44)
 {
    foutput("[WAR]  Expected %"PRIu64" bytes but found %"PRIu64" on disc...patching data.\n",
    info->numbytes, info->file_size-44);
    info->numbytes = info->file_size - 44;
  }

  calc_info(info);
  return(0);
}

int audio_open(fileinfo_t* info)
{
  FLAC__StreamDecoderInitStatus result;

  info->audio=malloc(sizeof(audio_input_t));
  if (info->type==AFMT_WAVE)
    {
      info->audio->fp=fopen(info->filename,"rb");
      if (info->audio->fp==0)
        {
          return(1);
        }
      info->file_size = read_file_size(info->audio->fp);
      fseek(info->audio->fp,44,SEEK_SET);
      info->audio->bytesread=0;
    }
  else if (info->type==AFMT_FLAC)
    {
      info->audio->flac=FLAC__stream_decoder_new();
      info->audio->n=0;
      info->audio->eos=0;

      if (info->audio->flac==NULL)
        EXIT_ON_RUNTIME_ERROR_VERBOSE("ERR: Fatal error - could not create FLAC decoder\n")

        /* Legacy code: see above
        *    FLAC__file_decoder_set_filename(info->audio->flac,info->filename);
        *   FLAC__file_decoder_set_client_data(info->audio->flac,(void*)info);
        *    FLAC__file_decoder_set_write_callback(info->audio->flac,flac_write_callback);
        *    FLAC__file_decoder_set_error_callback(info->audio->flac,flac_error_callback);
        *    FLAC__file_decoder_set_metadata_callback(info->audio->flac,flac_metadata_callback); */

        result=/*FLAC__StreamDecoderInitStatus*/ FLAC__stream_decoder_init_file  	(
              /*FLAC__StreamDecoder *  */ 	 info->audio->flac,
              /*char * */ 	info->filename,
              /*FLAC__StreamDecoderWriteCallback */ 	flac_write_callback,
              /*FLAC__StreamDecoderMetadataCallback */ 	flac_metadata_callback,
              /*FLAC__StreamDecoderErrorCallback  */	flac_error_callback,
              (void *) info
            );

      if (result!=FLAC__STREAM_DECODER_INIT_STATUS_OK)
        {

          FLAC__stream_decoder_delete(info->audio->flac);
          EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Failed to initialise FLAC decoder\n");
        }


      if (!FLAC__stream_decoder_process_until_end_of_metadata(info->audio->flac))
        {
          FLAC__stream_decoder_delete(info->audio->flac);
          EXIT_ON_RUNTIME_ERROR_VERBOSE( "[ERR]  Failed to read metadata from FLAC file\n")

        }
    }
  return(0);
}

// Read numbytes of audio data, and convert it to DVD byte order
int audio_read(fileinfo_t* info, uint8_t* buf, int count)
{
  uint32_t i;
  uint8_t x;
  int n=0;
  int bytesread;
  FLAC__bool result;

  count-=(count%info->sampleunitsize);
  if (count%info->sampleunitsize)
    {
      foutput("Requested %d bytes, sampleunitsize=%d\n",count,info->sampleunitsize);
      fflush(stdout);
    }

  if (info->type==AFMT_WAVE)
    {
      n=fread(buf,1,count,info->audio->fp);
      if (info->audio->bytesread+n > info->numbytes)
        {
          n=info->numbytes-info->audio->bytesread;
        }
      info->audio->bytesread+=n;
      bytesread=n;
      while ((info->audio->bytesread < info->numbytes) && (bytesread < count))
        {
          n=fread(&buf[bytesread],1,count-bytesread,info->audio->fp);
          if (info->audio->bytesread+n > info->numbytes)
            {
              n=info->numbytes-info->audio->bytesread;
            }
          info->audio->bytesread+=n;
          bytesread+=n;
        }
      n=bytesread;
    }
  else if (info->type==AFMT_FLAC)
    {
      while ((info->audio->n < count) && (info->audio->eos==0))
        {
          result=FLAC__stream_decoder_process_single(info->audio->flac);

          if (result==0)
            EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Fatal error decoding FLAC file\n")

            if (FLAC__stream_decoder_get_state(info->audio->flac)==FLAC__STREAM_DECODER_END_OF_STREAM)
              {
                info->audio->eos=1;
              }
        }
      if (info->audio->n >= count)
        {
          n=count;
          memcpy(buf,info->audio->buf,count);
          memmove(info->audio->buf,&(info->audio->buf[count]),info->audio->n-count);
          info->audio->n-=count;
        }
      else
        {
          n=info->audio->n;
          memcpy(buf,info->audio->buf,info->audio->n);
          info->audio->n=0;
        }
    }

  if (n % info->sampleunitsize)
    {
      foutput("%s", "[WAR]   Padding track with 1 sample\n");
      fflush(stdout);
      memcpy(&buf[n],&buf[n-6],6);
      n+=6;
    }

  // Convert little-endian WAV samples to big-endian MPEG LPCM samples
  if (info->bitspersample==16)
    {
      for (i=0;i<count;i+=2)
        {
          x=buf[i+1];
          buf[i+1]=buf[i];
          buf[i]=x;
        }
    }
  else if (info->bitspersample==24)
    {
      if (info->channels==1)
        {
          /* 24-bit mono samples are packed as follows:

                0  1  2  3  4  5
          WAV: 01 23 45 12 34 56
          AOB: 45 23 56 34 01 12

          */
          for (i=0;i<count;i+=6)
            {
              x=buf[i];
              buf[i]=buf[i+2];
              buf[i+2]=buf[i+5];
              buf[i+5]=buf[i+3];
              buf[i+3]=buf[i+4];
              buf[i+4]=x;
            }
        }
      else
        {
          /* 24-bit Stereo samples are packed as follows:

                0  1  2  3  4  5  6  7  8  9 10 11
          WAV: 01 23 45 bf 60 8c 67 89 ab b7 d4 e3
          AOB: 45 23 8c 60 ab 89 e3 d4 01 bf 67 b7

          */

          for (i=0;i<count;i+=12)
            {
              x=buf[i];
              buf[i]=buf[i+2];
              buf[i+2]=buf[i+5];
              buf[i+5]=buf[i+7];
              buf[i+7]=buf[i+10];
              buf[i+10]=buf[i+6];
              buf[i+6]=buf[i+11];
              buf[i+11]=buf[i+9];
              buf[i+9]=buf[i+3];
              buf[i+3]=buf[i+4];
              buf[i+4]=buf[i+8];
              buf[i+8]=x;
            }
        }
    }
  else
    {
      /* 20-bit stereo samples are packed as follows:

      Packet 0: 1980 bytes
      Packets 1-: 2000 bytes

      Stored similarly to 24-bit:

      4 samples, most significant 16 bits of each sample first, in big-endian
      order, followed by 2 bytes containing the least-significant 4 bits of
      each sample.

      I'm guessing  that  20-bits are stored in the most-significant
      20-bits of the 24.
      */

      // FIX: Handle 20-bit audio and maybe convert other formats.
      foutput("[ERR]  %d bit audio is not supported\n",info->bitspersample);
      clean_exit(EXIT_SUCCESS, DEFAULT);
    }
  return(n);
}

int audio_close(fileinfo_t* info)
{
  if (info->type==AFMT_WAVE)
    {
      fclose(info->audio->fp);
    }
  else if (info->type==AFMT_FLAC)
    {
      FLAC__stream_decoder_delete(info->audio->flac);
    }
  free(info->audio);
  return(0);
}

void uint32_copy(uint8_t* buf, uint32_t x)
{
  buf[0]=(x>>24)&0xff;
  buf[1]=(x>>16)&0xff;
  buf[2]=(x>>8)&0xff;
  buf[3]=x&0xff;
}

void uint16_copy(uint8_t* buf, uint16_t x)
{
  buf[0]=(x>>8)&0xff;
  buf[1]=x&0xff;
}

